"""Common types."""
from datetime import date, datetime
from enum import Enum, unique
from typing import Any, Dict, Generic, List, Optional, Sequence, TypeVar

from pydantic import BaseModel
from pydantic.generics import GenericModel
from regal import Address, ContactId
from regal.common import BaseId

T = TypeVar("T", bound=BaseModel)


class VerificationId(BaseId):
    """Verification id."""


class Pagination(BaseModel):
    """Token pagination model."""

    next_page_token: str = ""
    items_left: Optional[int] = None


class PaginatedList(GenericModel, Generic[T]):
    """Token paginated results."""

    results: List[T]
    pagination: Optional[Pagination] = None


@unique
class VerificationStatus(str, Enum):
    """Verification status."""

    UNKNOWN = "unknown"
    CREATED = "created"
    INITIALIZED = "initialized"
    ACCESSED = "accessed"
    PENDING = "pending"
    SUCCEEDED = "succeeded"
    FAILED = "failed"
    CANCELED = "canceled"


class Event(BaseModel):
    """Verification event."""

    event_datetime: datetime
    type: str
    method: Optional[str] = None
    details: Dict[str, Any] = {}


class Verification(BaseModel):
    """Verification info."""

    verification_id: VerificationId
    contact_id: ContactId
    create_datetime: datetime
    close_datetime: Optional[datetime] = None
    status: VerificationStatus = VerificationStatus.UNKNOWN

    events: Sequence[Event] = []
    next_events: Sequence[Event] = []


class ContactData(BaseModel):
    """Contact data to be verified."""

    name: Optional[str] = None
    organization: Optional[str] = None
    birthdate: Optional[date] = None
    company_registration_number: Optional[str] = None
    address: Optional[Address] = None


class VerificationResult(BaseModel):
    """Contact data verification result."""

    verify_datetime: datetime
    method: str
    data: ContactData
    contact_history_id: str

    # True - contact field passed verfication
    # False - contact field failed verification
    # None - contact field was not verified
    name: Optional[bool] = None
    organization: Optional[bool] = None
    birthdate: Optional[bool] = None
    company_registration_number: Optional[bool] = None
    address: Optional[bool] = None

    @property
    def success(self) -> bool:
        """Return True if contact data verification didn't fail."""
        return False not in (
            self.name,
            self.organization,
            self.birthdate,
            self.company_registration_number,
            self.address,
        )


class ContactRepresentative(BaseModel):
    """Contact representative."""

    organization: str
    address: Address
